package com.framsticks.communication;

import com.framsticks.communication.queries.ApplicationRequest;
import com.framsticks.communication.queries.RegistrationRequest;
import com.framsticks.communication.queries.UseRequest;
import com.framsticks.communication.queries.VersionRequest;
import com.framsticks.communication.util.LoggingStateCallback;
import com.framsticks.params.ListSource;
import com.framsticks.util.*;
import com.framsticks.util.dispatching.AtOnceDispatcher;
import com.framsticks.util.dispatching.Dispatcher;
import com.framsticks.util.dispatching.Dispatching;
import com.framsticks.util.lang.Pair;
import com.framsticks.util.lang.Strings;
import org.apache.log4j.Logger;

import java.io.IOException;
import java.net.Socket;
import java.net.SocketException;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import com.framsticks.util.dispatching.RunAt;

/**
 * @author Piotr Sniegowski
 */
public class ClientConnection extends Connection {

	private final static Logger log = Logger.getLogger(ClientConnection.class);

	protected final Map<String, Subscription<?>> subscriptions = new HashMap<>();

	/**
	 * @param connectedFunctor the connectedFunctor to set
	 */
	public void setConnectedFunctor(StateFunctor connectedFunctor) {
		this.connectedFunctor = connectedFunctor;
	}


	protected StateFunctor connectedFunctor;

	@Override
	protected void joinableStart() {
		try {
			log.info("connecting to " + address);

			socket = new Socket(hostName, port);

			socket.setSoTimeout(500);

			log.info("connected to " + hostName + ":" + port);
			connected = true;
			runThreads();

			connectedFunctor.call(null);
		} catch (SocketException e) {
			log.error("failed to connect: " + e);
			connectedFunctor.call(e);
		} catch (IOException e) {
			log.error("buffer creation failure");
			connectedFunctor.call(e);
			// close();
		}
	}


	private static abstract class InboundMessage {
		protected String currentFilePath;
		protected List<String> currentFileContent;
		protected final List<File> files = new ArrayList<File>();

		public abstract void eof();

		protected void initCurrentFile(String path) {
			currentFileContent = new LinkedList<String>();
			currentFilePath = path;
		}
		protected void finishCurrentFile() {
			if (currentFileContent == null) {
				return;
			}
			files.add(new File(currentFilePath, new ListSource(currentFileContent)));
			currentFilePath = null;
			currentFileContent= null;
		}

		public abstract void startFile(String path);

		public void addLine(String line) {
			assert line != null;
			assert currentFileContent != null;
			currentFileContent.add(line.substring(0, line.length() - 1));
		}

		public List<File> getFiles() {
			return files;
		}
	}

	private static class EventFire extends InboundMessage {
		public final Subscription<?> subscription;

		private EventFire(Subscription<?> subscription) {
			this.subscription = subscription;
		}

		public void startFile(String path) {
			assert path == null;
			initCurrentFile(null);
		}

		@Override
		public void eof() {
			finishCurrentFile();

			subscription.dispatchCall(getFiles());
		}
	}

	private static class SentQuery<C> extends InboundMessage {
		Request request;
		ResponseCallback<? extends C> callback;
		Dispatcher<C> dispatcher;

		public void startFile(String path) {
			finishCurrentFile();
			if (path == null) {
				assert request instanceof ApplicationRequest;
				path = ((ApplicationRequest)request).getPath();
			}
			initCurrentFile(path);
		}

		public void eof() {
			finishCurrentFile();
			//no-operation
		}

		@Override
		public String toString() {
			return request.toString();
		}

		public void dispatchResponseProcess(final Response response) {
			Dispatching.invokeLaterOrNow(dispatcher, new RunAt<C>() {
				@Override
				public void run() {
					callback.process(response);
				}
			});
		}
	}
	private Map<Integer, SentQuery<?>> queryMap = new HashMap<>();


	protected final String hostName;
	protected final int port;

	private static Pattern addressPattern = Pattern.compile("^([^:]*)(:([0-9]+))?$");

	public ClientConnection(String address) {
		super(address);
		Matcher matcher = addressPattern.matcher(address);
		if (!matcher.matches()) {
			log.fatal("invalid address: " + address);
			hostName = null;
			port = 0;
			return;
		}
		hostName = matcher.group(1);
		port = matcher.group(3) != null ? Integer.parseInt(matcher.group(3)) : 9009;
	}

	private SentQuery<?> currentlySentQuery;


	public <C extends Connection> void send(Request request, ResponseCallback<C> callback) {
		//TODO RunAt
		send(request, AtOnceDispatcher.getInstance(), callback);
	}

	public <C> void send(Request request, Dispatcher<C> dispatcher, ResponseCallback<? extends C> callback) {

		if (!isConnected()) {
			log.fatal("not connected");
			return;
		}
		final SentQuery<C> sentQuery = new SentQuery<C>();
		sentQuery.request = request;
		sentQuery.callback = callback;
		sentQuery.dispatcher = dispatcher;

		senderThread.invokeLater(new RunAt<Connection>(){
			@Override
			public void run() {
				Integer id;
				synchronized (ClientConnection.this) {

					while (!(requestIdEnabled || currentlySentQuery == null)) {
						try {
							ClientConnection.this.wait();
						} catch (InterruptedException ignored) {
							break;
						}
					}
					if (requestIdEnabled) {
						queryMap.put(nextQueryId, sentQuery);
						id = nextQueryId++;
					} else {
						currentlySentQuery = sentQuery;
						id = null;
					}
				}
				String command = sentQuery.request.getCommand();
				StringBuilder message = new StringBuilder();
				message.append(command);
				if (id != null) {
					message.append(" ").append(id);
				}
				sentQuery.request.construct(message);
				String out = message.toString();

				output.println(out);
				log.debug("sending query: " + out);

			}
		});
		/*
		synchronized (this) {
			log.debug("queueing query: " + query);
			queryQueue.offer(sentQuery);
			notifyAll();
		}
		*/
	}


	@Override
	public String toString() {
		return "client connection " + address;
	}

	public <C> void subscribe(final String path, final Dispatcher<C> dispatcher, final SubscriptionCallback<? extends C> callback) {
		send(new RegistrationRequest().path(path), new ResponseCallback<Connection>() {
			@Override
			public void process(Response response) {
				if (!response.getOk()) {
					log.error("failed to register on event: " + path);
					callback.subscribed(null);
					return;
				}
				assert response.getFiles().isEmpty();
				Subscription<C> subscription = new Subscription<C>(ClientConnection.this, path, response.getComment(), dispatcher);
				log.debug("registered on event: " + subscription);
				synchronized (subscriptions) {
					subscriptions.put(subscription.getRegisteredPath(), subscription);
				}
				subscription.setEventCallback(callback.subscribed(subscription));
				if (subscription.getEventCallback() == null) {
					log.info("subscription for " + path + " aborted");
					subscription.unsubscribe(new LoggingStateCallback<C>(log, "abort subscription"));
				}
			}
		});
	}

	public void negotiateProtocolVersion(StateFunctor stateFunctor) {
		protocolVersion = -1;
		sendQueryVersion(1, stateFunctor);
	}

	public void sendQueryVersion(final int version, final StateFunctor stateFunctor) {
		send(new VersionRequest().version(version), new StateCallback<Connection>() {
			@Override
			public void call(Exception e) {
				if (e != null) {
					log.fatal("failed to upgrade protocol to version: " + version);
					return;
				}
				protocolVersion = version;
				if (version < 4) {
					/** it is an implicit loop here*/
					sendQueryVersion(version + 1, stateFunctor);
					return;
				}
				send(new UseRequest().feature("request_id"), new StateCallback<Connection>() {
					@Override
					public void call(Exception e) {
						requestIdEnabled = e == null;
						/*
						synchronized (ClientConnection.this) {
							ClientConnection.this.notifyAll();
						}
						*/
						if (!requestIdEnabled) {
							log.fatal("protocol negotiation failed");
							stateFunctor.call(new Exception("protocol negotiation failed", e));
							return;
						}
						stateFunctor.call(null);
					}
				});

			}
		});
	}


	private synchronized SentQuery<?> fetchQuery(Integer id, boolean remove) {
		if (id == null) {
			if (requestIdEnabled) {
				return null;
			}
			SentQuery<?> result = currentlySentQuery;
			if (remove) {
				currentlySentQuery = null;
				notifyAll();
			}
			return result;
		}
		if (queryMap.containsKey(id)) {
			SentQuery<?> result = queryMap.get(id);
			if (remove) {
				queryMap.remove(id);
			}
			return result;
		}
		return null;
	}

	private int nextQueryId = 0;

	protected void processMessage(InboundMessage inboundMessage) {
		if (inboundMessage == null) {
			log.error("failed to use any inbound message");
			return;
		}

		String line;
		while (!(line = getLine()).startsWith("eof")) {
			// log.debug("line: " + line);
			inboundMessage.addLine(line);
		}
		inboundMessage.eof();
	}

	protected void processEvent(String rest) {
		Matcher matcher = eventPattern.matcher(rest);
		if (!matcher.matches()) {
			log.error("invalid event line: " + rest);
			return;
		}
		Subscription<?> subscription = subscriptions.get(matcher.group(1));
		if (subscription == null) {
			log.error("non subscribed event: " + matcher.group(1));
			return;
		}
		EventFire event = new EventFire(subscription);
		event.startFile(null);
		processMessage(event);
	}


	protected void processMessageStartingWith(String line) {
		Pair<String, String> command = Strings.splitIntoPair(line, ' ', "\n");
		if (command.first.equals("event")) {
			processEvent(command.second);
			return;
		}
		Pair<Integer, String> rest = parseRest(command.second);

		if (command.first.equals("file")) {
			SentQuery<?> sentQuery = fetchQuery(rest.first, false);
			sentQuery.startFile(rest.second);
			processMessage(sentQuery);
			return;
		}

		SentQuery<?> sentQuery = fetchQuery(rest.first, true);
		if (sentQuery == null) {
			return;
		}
		log.debug("parsing response for request " + sentQuery);

		sentQuery.dispatchResponseProcess(new Response(command.first.equals("ok"), rest.second, sentQuery.getFiles()));
	}

	@Override
	protected void receiverThreadRoutine() {
		while (connected) {
			try {
				processMessageStartingWith(getLine());
			} catch (Exception e) {
				break;
			}
		}
	}


}
