package com.framsticks.gui;

import com.framsticks.communication.Connection;
import com.framsticks.communication.Subscription;
import com.framsticks.communication.util.LoggingStateCallback;
import com.framsticks.core.Instance;
import com.framsticks.core.ListChange;
import com.framsticks.core.Path;
import com.framsticks.gui.controls.ValueControl;
import com.framsticks.gui.view.TreeCellRenderer;
import com.framsticks.params.AccessInterface;
import com.framsticks.params.CompositeParam;
import com.framsticks.params.FramsClass;
import com.framsticks.params.types.*;
import com.framsticks.remote.*;
import com.framsticks.util.lang.Casting;
import com.framsticks.util.swing.TooltipConstructor;
import com.framsticks.util.dispatching.Future;
import com.framsticks.util.Logging;
import com.framsticks.util.StateFunctor;
import org.apache.log4j.Logger;
import com.framsticks.util.dispatching.RunAt;

import javax.swing.tree.DefaultMutableTreeNode;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import static com.framsticks.util.lang.Containers.filterInstanceof;
import com.framsticks.util.swing.TreeNodeUtils;

/**
 * @author Piotr Sniegowski
 */
@SuppressWarnings("serial")
public class TreeNode extends DefaultMutableTreeNode implements NodeListener {

	private static final Logger log = Logger.getLogger(TreeNode.class.getName());

	Panel panel;

	final protected Frame frame;
	final protected InstanceAtFrame instanceAtFrame;

	final protected Map<EventParam, Subscription<?>> userSubscriptions = new HashMap<>();
	protected Map<ValueControl, Object> localChanges = null;
	protected String tooltip;
	protected String name;
	protected final String paramId;
	protected String iconName;
	protected Path path;

	public TreeNode(InstanceAtFrame instanceAtFrame, final Path path) {
		this.frame = instanceAtFrame.getFrame();
		assert frame.isActive();
		this.paramId = path.getLastElement();
		this.name = this.paramId;
		log.debug("creating treenode " + name + ": " + path);
		this.path = path;
		this.instanceAtFrame = instanceAtFrame;

		iconName = TreeCellRenderer.findIconName(name, path.getTextual());
		tooltip = "?";
		path.getInstance().invokeLater(new RunAt<Instance>() {
			@Override
			public void run() {
				updateDescriptions(path);
			}
		});
	}

	public void fetch(final Path p) {
		assert p.getInstance().isActive();
		log.debug("fetching: " + p);
		p.getInstance().fetchValues(p, new StateFunctor() {
			@Override
			public void call(Exception e) {
				//TODO removing should stay here
				// reactForFetchResult(p, e);
			}
		});
	}

	protected void reactForFetchResult(final Path p, Exception e) {
		assert p.getInstance().isActive();
		if (Logging.log(log, "fetch", TreeNode.this, e)) {
			frame.invokeLater(new RunAt<Frame>() {
				@Override
				public void run() {
					log.debug("removing node from tree: " + p);
					frame.treeModel.removeNodeFromParent(TreeNode.this);
				}
			});
			return;
		}
		updateChildren(p);
		frame.invokeLater(new RunAt<Frame>() {
			@Override
			public void run() {
				//TODO maybe this should be called from some better place
				useOrCreatePanel();
			}
		});
	}

	protected void postUpdatePath(final Path newPath) {
		assert !frame.isActive();
		/** TODO those two actions could be merged into single closure */
		frame.invokeLater(new RunAt<Frame>() {
			@Override
			public void run() {
				updatePath(newPath);
			}
		});
		updateDescriptions(newPath);
	}

	protected void updatePath(Path newPath) {
		assert frame.isActive();
		if (!path.isResolved()) {
			path = newPath;
			log.debug("updated treenode's path: " + path);
			return;
		}
		if (path.matches(newPath)) {
			return;
		}
		log.debug("removing all children due to path update");
		this.removeAllChildren();
		frame.treeModel.nodeStructureChanged(this);
	}

	public Iterable<TreeNode> childrenIterable() {
		return filterInstanceof(TreeNodeUtils.children(TreeNode.this, frame.treeModel), TreeNode.class);
	}

	/** Update children, by removing non existing and adding new ones.
	 */
	protected void updateChildren(final Path p) {
		assert p.getInstance().isActive();
		log.debug("updating children of " + this);
		AccessInterface access = p.getInstance().bindAccess(p.getTop());
		if (access == null) {
			return;
		}
		final List<Path> childrenPaths = new LinkedList<Path>();
		/**Prepare path for each child.*/
		for (CompositeParam param : filterInstanceof(access.getParams(), CompositeParam.class)) {
			childrenPaths.add(p.appendParam(param).tryFindResolution());
		}

		/**If some child were found, update in frame context.*/
		if (childrenPaths.size() > 0) {
			frame.invokeLater(new RunAt<Frame>() {
				@Override
				public void run() {
					// TreePath treePath = frame.startChange();
					updatePath(p);
					Map<String, TreeNode> current = new HashMap<String, TreeNode>();
					for (TreeNode n : childrenIterable()) {
						current.put(n.path.getLastElement(), n);
					}
					assert current.size() == TreeNode.this.getChildCount();
					assert TreeNode.this.getChildCount() == frame.treeModel.getChildCount(TreeNode.this);
					log.trace("current " + TreeNode.this.getChildCount() + " children: " + current.values());
					for (Path childPath : childrenPaths) {
						String e = childPath.getLastElement();
						if (current.containsKey(e)) {
							current.remove(e);
							continue;
						}
						log.debug("update: treenode for " + p);
						TreeNode childNode = new TreeNode(instanceAtFrame, childPath);

						frame.addNode(childNode, TreeNode.this);
					}
					for (TreeNode n : current.values()) {
						frame.treeModel.removeNodeFromParent(n);
					}
					log.debug("updated children of " + TreeNode.this + ": " + TreeNode.this.getChildCount());
				}
			});
		} else {
			log.debug("no children in " + TreeNode.this);
		}
	}

	public void select() {
		final Path p = path;

		p.getInstance().invokeLater(new RunAt<Instance>() {
			@Override
			public void run() {
				final Path updated = (p.isResolved()) ? p : p.tryFindResolution();
				if (updated.isResolved()) {
					Logging.log(log, "update", updated, null);
					fetch(updated);
					postUpdatePath(updated);
					return;
				}
				p.getInstance().resolve(updated, new Future<Path>() {
					@Override
					public void result(final Path result, Exception e) {
						if (Logging.log(log, "resolve and select", TreeNode.this, e)) {
							return;
						}

						fetch(result);
						postUpdatePath(result);
					}
				});
			}
		});

	}

	@Override
	public String toString() {
		return path.toString();
	}

	public final Panel getPanel() {
		assert frame.isActive();
		return panel;
	}

	protected void updateDescriptions(Path p) {
		assert p.getInstance().isActive();

		if (!p.isResolved()) {
			return;
		}
		AccessInterface access = p.getInstance().bindAccess(p);
		if (access == null) {
			return;
		}

		final String tooltip = new TooltipConstructor()
			.append("frams", access.getId())
			.append("java", p.getTopObject().getClass().getCanonicalName())
			.append("access", access.getClass().getSimpleName())
			.append("name", name)
			.append("id", paramId)
			.append("object", Integer.toHexString(System.identityHashCode(this)))
			.build()
			;

		StringParam nameParam = Casting.tryCast(StringParam.class, access.getParam("name"));
		final String name = (nameParam != null ? access.get(nameParam, String.class) : path.getTop().getParam().getId());

		frame.invokeLater(new RunAt<Frame>() {
			@Override
			public void run() {
				TreeNode.this.tooltip = tooltip;
				TreeNode.this.name = name;

				log.debug("updated descriptions for " + path + ": " + name);
			}
		});
	}

/*
	public void updateData() {
		assert browser.isActive();
		final Node node = getNode();
		browser.manager.invokeLater(new Runnable() {
			@Override
			public void run() {
				node.fetchValues(new StateFunctor() {
					@Override
					public void call(Exception e) {
						if (e != null) {
							return;
						}
						browser.invokeLater(new Runnable() {
							@Override
							public void run() {
								assert browser.isActive();
								if (panel.getCurrentTreeNode() == TreeNode.this) {
									panel.refreshComponents();
								}

								browser.tree.repaint();
								panel.refreshComponents();
							}
						});
					}
				});
			}
		});
	}
*/

	public void showPanel() {
		assert frame.isActive();
		assert panel != null;
		frame.showPanel(panel);
	}

	public void fillPanelWithValues() {
		assert frame.isActive();
		if (panel == null) {
			return;
		}
		final Path p = path;
		assert p.isResolved();
		panel.setCurrentTreeNode(this);
		p.getInstance().invokeLater(new RunAt<Instance>() {
			@Override
			public void run() {
				AccessInterface access = p.getInstance().bindAccess(p);
				panel.pullValuesFromLocalToUser(access);

				frame.invokeLater(new RunAt<Frame>() {
					@Override
					public void run() {
						showPanel();
					}
				});
			}
		});

	}

	public void useOrCreatePanel() {
		assert frame.isActive();
		if (panel != null) {
			log.trace("panel is already attached: " + path);
			fillPanelWithValues();
			return;
		}
		if (!path.isResolved()) {
			log.trace("path is not resolved: " + path);
			return;
		}

		CompositeParam param = path.getTop().getParam();
		panel = instanceAtFrame.findPanel(param.computeAccessId());
		if (panel != null) {
			log.debug("found prepared panel for: " + path);
			fillPanelWithValues();
			return;
		}
		final Path p = path;
		log.debug("preparing panel: " + p);
		p.getInstance().invokeLater(new RunAt<Instance>() {
			@Override
			public void run() {
				assert p.getInstance().isActive();
				final CompositeParam param = p.getTop().getParam();
				final FramsClass framsClass = p.getInstance().getInfoFromCache(param.getContainedTypeName());
				frame.invokeLater(new RunAt<Frame>() {
					@Override
					public void run() {
						panel = instanceAtFrame.preparePanel(param, framsClass);
						fillPanelWithValues();
					}
				});
			}
		});
	}

	public String getTooltip() {
		assert frame.isActive();
		return tooltip;
	}




	/*public void subscribe(final EventParam eventParam) {
		assert browser.isActive();
		if (hasSubscribed(eventParam)) {
			log.error(eventParam + " is already subscribed for " + this);
			return;
		}
		Node node = getNode();
		node.getConnection().subscribe(node.getPath() + "/" + eventParam.getId(), new SubscriptionCallback() {
			@Override
			public EventCallback subscribed(final Subscription subscription) {
				if (subscription == null) {
					log.error("subscription failed");
					return null;
				}
				if (hasSubscribed(eventParam)) {
					//abort subscription
					return null;
				}
				userSubscriptions.put(eventParam, subscription);
				log.debug("subscription succeeded: " + subscription);
				subscription.setDispatcher(browser);
				return new EventCallback() {
					@Override
					public void call(SourceInterface content) {
						assert browser.isActive();
						log.info("event " + subscription + " occurred");
					}
				};
			}
		});

	}*/

	public boolean hasSubscribed(EventParam param) {
		assert frame.isActive();
		return userSubscriptions.containsKey(param);
	}

	public void unsubscribe(EventParam eventParam) {
		assert frame.isActive();
		if (!hasSubscribed(eventParam)) {
			log.error("could not unsubscribe from " + eventParam);
			return;
		}
		userSubscriptions.get(eventParam).unsubscribe(new LoggingStateCallback<Connection>(log, "unsubscribed " + eventParam));
		userSubscriptions.remove(eventParam);
	}

/*

	@Override
	public void onChildChange(final Child child, ListChange.Action action) {
		assert browser.manager.isActive();

		switch (action) {
			case Remove: {
				Dispatching.invokeDispatch(browser, browser.manager, new Runnable() {
					@Override
					public void run() {
						assert browser.manager.isActive();
						final TreeNode treeNode = (TreeNode) child.getUserObject();
						if (treeNode == null) {
							log.error("child " + child + " had no tree node attached");
							return;
						}
						browser.invokeLater(new Runnable() {
							@Override
							public void run() {
								assert browser.isActive();
								TreePath path = browser.startChange();
								//assert treeNode.getParent() == TreeNode.this;
								if (treeNode.getParent() != null) {
									browser.treeModel.removeNodeFromParent(treeNode);
								}
								//remove(treeNode);
								browser.markNodeChanged(TreeNode.this, path);
							}
						});
					}
				});
				break;
			}
		}

	}
*/

	public String getName() {
		return name;
	}

	public String getIconName() {
		return iconName;
	}

	@Override
	public void onUpgrade(Path path) {
	}

	@Override
	public void onChange(Path path) {
	}

	@Override
	public void onChildChange(Path path, ListChange.Action action) {
	}

	public final Frame getFrame() {
		return frame;
	}

	public boolean changeValue(ValueControl component, Object newValue) {
		log.debug("changing value of " + component + " to '" + newValue + "'");

		if (localChanges == null) {
			localChanges = new HashMap<ValueControl, Object>();
		}
		localChanges.put(component, newValue);

		return true;
	}

	public Path getInstancePath() {
		assert frame.isActive();
		return path;
	}

	public void pushLocalChanges() {
		assert frame.isActive();
		if (localChanges == null) {
			return;
		}
		final Map<ValueControl, Object> changes = localChanges;
		localChanges = null;
		instanceAtFrame.getInstance().invokeLater(new RunAt<Instance>() {
			@Override
			public void run() {
				for (Map.Entry<ValueControl, Object> e : changes.entrySet()) {
					final ValueControl key = e.getKey();
					final Path p = path;
					instanceAtFrame.getInstance().storeValue(p, e.getKey().getParam(), e.getValue(), new StateFunctor() {
						@Override
						public void call(Exception e) {
							changes.remove(key);
							if (!changes.isEmpty()) {
								return;
							}
							log.debug("applied changes for: " + p);
							frame.invokeLater(new RunAt<Frame>() {
								@Override
								public void run() {
									fillPanelWithValues();
								}
							});
						}
					});
				}
			}
		});
	}
}
